/* (c) Joerg Schulenburg 2000-2007

 this is part of spin-package

 it is for division of matrix multiplication 
 in parallelizable block matrix multiplication
 
    b = H * a    (a,b vectors, H matrix)
    
    b[i] = sum_j H[i,j] * a[j]
    
 blocks can be multiplicated step by step on one system or
 they can be multiplicated on different processors/systems
 
 but you have to pay additional network/harddisk traffic
 
 if b_num is one you have only one block (similar to old version)


 */

#ifndef _HBLOCK_H
#define _HBLOCK_H

#include "config.h"

#ifndef B_NUM
#define B_NUM 4 /* num of compute blocks */
#endif

#ifndef F_NUM
#define F_NUM B_NUM      /* later moved to config.h, can be modified */
#endif

/* #include "defs.h" */

extern unsigned long b_ofs[B_NUM];	/* start index 0..n1-1 of block */
extern unsigned long b_len[B_NUM];	/* block len */
extern unsigned long b_hnz[B_NUM];	/* NumNonZero-H block entries */
/* this is for partly stored files, ToDo: remove or replace */
extern unsigned long b_snz[B_NUM];	/* stored NonZero-H block entries */
extern unsigned long b_sn1[B_NUM];	/* stored block lines */
extern int b_bits;			/* b_len[0..n-2]=1l<<b_bits */
extern unsigned long b_mask;			/* ~((~0)<<b_bits) */

/* calculate blocknumber (0..B_NUM-1) from idx (0..n1-1) */
/* used for store blocks of hamilton (future) and converted relikts */
static inline int b_getnum(unsigned long idx){	/* do it fast as possible! */
#if   B_NUM <= 1    /* avoid error if H is not splitted */
  if (idx) ; /* dummy */ return 0;
#elif B_NUM == 2    /* performance boost for two blocks */
  return ((idx >= b_ofs[1]) ? 1 : 0);
#elif B_NUM == 4    /* performance boost for four blocks, better divide? */
  /* return (idx/b_ofs[1]); // slower and dangerous for future changes! */
  /* faster than if() on x86 */
  return ((idx >= b_ofs[2]) ? ((idx >= b_ofs[3]) ? 3 : 2) :
                              ((idx >= b_ofs[1]) ? 1 : 0) );
#else
  int a, m, b;       /* left border, middle, right border 2.26 Feb07 */
  for (a=0,m=B_NUM/2,b=B_NUM; a!=m; m=(a+b)/2) {
    if( idx >= b_ofs[m] ) a=m; else b=m;
  }
  return m;
#endif
}

/* initialization, n1 is whole matrix size */
int b_ini(unsigned long n1,int vvv);

#include "h_file.h"
extern struct thxy b_hbuf[B_NUM]; /* buffer for faster access */ 
extern int          b_hbl[B_NUM]; /* num valid data in buffer bufh */  
#endif
