/* 
 *  send signal n (by sending n*USR2 + USR1) to the running process
 */

#include <stdlib.h>
#include <stdio.h>

/* next two includes needed for getpid(), outcommend on problems */
#include <sys/types.h>
#include <unistd.h>

/* outcomment next include, if your system does not support it */
#ifndef NoSIGNAL     /* may be defined by configure in Makefile */
#include <signal.h>  /* broken for SLES9 g++-4.1 at IA-64 */
/*  linux: _SIGNAL_H, OSF1: _SYS_SIGNAL_H_ */
#endif

/* simply outcomment next line, if you have problems with time.h */
#include <time.h>	/* defines time() and clock() */
/* time.h often defines ulong (therefore we use ULong) */

#include "error.h"
#include "sighand.h"

/* export: int sig1; int ini_handler(); */
volatile int sig1;	/* set by signal handler */
int stkptr=0;		/* stack mit stk[stkptr].p="pg-name",.i=counter */
struct struct_stk stk[20];	/* max. Tiefe stkptr=20 */


unsigned long starttime=0l;
#if defined(_TIME_H) || defined(_TIME_H_)
/* for the case that time.h was not included, skip time code
 *  linux: _TIME_H, OSF1: _TIME_H_,_SYS_SIGNAL_H_
 */
long itime(){	/* ProzessZeit=clock()/CLOCKS_PER_SEC unixzeit=time */
  time_t i;time(&i);if(starttime==0l)starttime=i;return(long)(i-starttime);}
long iclock(){ return(long)clock();}
#else
#warning "no time.h, time code is dropped"
long itime(){ return 0; }
long iclock(){ return 0; }
#endif

/*
+++TeX+++
 This part of the program allows the user to control
 the running process a bit. You are able to stop some
 functions in a controled manner or get further information,
 what the running process is doing.
	
 This is done by using unix signals (SIGUSR1 and SIGUSR2).
 On unix systems you can do it via the kill command from the command line,
 which are usually used to stop or continue a unix process.
	
 in German (have to translate it):
 
	Ist die Prozess-Identifikations-Nummer (PID) zum Beispiel 1234, so
	kann mit {\bf kill -SIGUSR1 1234} oder {\bf kill -SIGUSR2 1234}
	eines der beiden Signale an den Prozess gesendet werden.
	Das Programm reagiert dann beispielsweise mit der Ausgabe
	einer Tabelle, die die gerade ablaufenden Funktionen und einen
	spezifischen Zaehlerstand ausgibt.
	
	Die Tabelle kann dann bei Start eines Unterprogramms erweitert
	und bei verlassen des Programms wieder verkleinert werden.
	Mit einer Kombination von $n$ Signalen SIGUSR2 und einem
	SIGUSR1 koennen auch kompliziertere Reaktionen vom Programm 
	erzeugt werden, da das Programm die Variable {\bf sig1}, die
	dann $n$ enthaelt, jederzeit auslesen kann.
---TeX---
*/

/* output stack */
void o_stack(){ int i;
   err_pg(0);	/* err-status */
   for(i=0;i<stkptr;i++){
     printf("#%d# %20s",i,stk[i].p); 
     if (stk[i].i) printf(" %lu",*(stk[i].i));
     printf("\n");
   }
}

#define PID_FILE "status"
#if defined(_SIGNAL_H) || defined(_SIGNAL_H_) || defined(_SYS_SIGNAL_H_)
void my_handler(int sig){ static int x=0; 
    switch(sig){
      case SIGUSR1: switch(x){
                     case 0: if(!sig1) o_stack(); break;
                     case 1: /*  */ break;} 
                    sig1=x; x=0; break;
      case SIGUSR2: x++; break;
      case SIGALRM: /* alarm(60); */ break; 
      default:      sig1=sig;printf("\nsignal(%d) received ",sig);}
    signal(sig,my_handler); /* redirect again */}
#endif

/* mehrfachaufruf moeglich! */
int ini_handler(int vvv){ /* better open and close_handler ??? */
   FILE *f1;
   /* store the own PID (Process IDentification number) in a file */
   /* can also be used, too communicate with the process (sh status) */
   f1=fopen(PID_FILE,"w");
   if(!f1)fprintf(stderr," error "__FILE__" %d: create "PID_FILE,__LINE__);
   if(f1){ if(vvv)printf(" write "PID_FILE);
#if defined(_UNISTD_H) || defined(_UNISTD_H_)
     fprintf(f1,"#!sh\n# get status of spin-pg (automaticly generated)\n"
                "# use: sh "PID_FILE"\nkill -USR1 %ld # or kill -SIGUSR1\n"
                "# writes to stdout (redirected to nohup.out or whatever)\n",
                (long)getpid() );
#else
#warning "no unistd.h, getpid code was dropped"
     fprintf(f1,"#!sh\necho \"sorry, getpid code was dropped\"\n");
#endif
     fclose(f1); }
   sig1=0;
#if defined(_SIGNAL_H) || defined(_SIGNAL_H_) || defined(_SYS_SIGNAL_H_)
   signal(SIGUSR1,my_handler);	/* fkt(x); x=0; (PG-Status) */
   signal(SIGUSR2,my_handler);  /* x=x+1; */
   signal(SIGALRM,my_handler);  /*  */
   /* if(vvv&16)alarm(60); */	/* SIgnal SIGALRM in 60s */
#else
#warning "no signal.h, communication code dropped"
#endif
   return 0;
}
