/*
 * Warning: char can be -128..127 or 0..255
 *  verbose|=8 gives more infos
 * ToDo:
 * - use nauty-package or cycle-tree or recursive search (simple)
 * - reading of k-vectors not good! probably quash T incl. k
 * - ini_aclass with cycle lengths! faster!
 * - test_twice() faster, if syms sorted! evl. auch bei get_config
 * - sort in full commutating groups (without products of that groups)
 *   analog k-Sym and ud-Sym (spare memory! + faster!?, LM:s1 etc.) 
 * - S=3/2 spins (simulated as 3 S=1/2) with different quantum numbers?
 * history:
 *  Apr02 - use nn instead NN, "less_sym"-bug found
 *  20.04.01 - sym.lm extended by an index for faster S1SYM (buggy, less sym)
 *  15.03.01 - test_prod(): bug fixed (not used),
 *             warn if old_generator is product of newer ones
 *  06.05.00 - test_product changed, bug if daten.sym is given
 *  08.04.00 - ini_ww IFww TBC added (undirected graph)
 *  26.01.99 - k=-5 used for skipping next 5 permutations
 */
#include "defs.h"	/* ULong etc. defined */
#include "sighand.h"
#include "modeldef.h"	/* NN,Nw defined */
/* #include "spins.h" */ /* iw,... declared, should be seperated in ww.h */
#include "symmetry.h"
#include <stdio.h>
#include <stdlib.h>

#define FILE_S1  "daten.sym"	/* input file */
#define FILE_S2  "daten.sy2"	/* all permutations P (Pi*Pj==Pj*Pi) */

/* -------------------------------------------------------------------- *
 *                       SYMMETRIEN
 * => see doc/symmetry.tex
 * 
 * ---------------------------------------------------------------------- */
int	w2w[NN][NN];		/* interaction matrix (faster) */ 

struct tsym sym;                        /* large object ??? */

/* since v2.20 +/- bondnumber (1...) is used instead of X*numpara+wj[] */
int ini_ww(){
  static int rdy=0;int i,j,idxold,idxnew,w1,ndiff=0;
  if (rdy) return 0; else rdy++;                     /* do it only once */
  printf(" size(sym)=%.2e\n",(double)sizeof(struct tsym));
  for(i=0;i<NN;i++)for(j=0;j<NN;j++) w2w[i][j]=0;

  for(w1=i=0;i<nw;i++){ /* create w2w-matrix for speedup (different bonds) */
    idxold= w2w[ iw[i][0] ][ iw[i][1] ];
    if (idxold) { if (!w1++) E1(" bond %d defined twice!",i); }
    idxnew= ndiff + 1; /* avoid 0 */
    for (j=0;j<i;j++)   /* search stored pairs for same */
      if (ww.wt[i] == ww.wt[j]  /* t hopping  */
       && ww.wj[i] == ww.wj[j]  /* J exchange */
       && ww.wv[i] == ww.wv[j]  /* V coulomb */
       && ww.wh[i] == ww.wh[j]  /* hole repulsion */
       IFww( && ww.w[i]==ww.w[j] ) /* phase related to direction R */
      ) { /* found same pair */
          idxnew = w2w[ iw[j][0] ][ iw[j][1] ] IFww(*ww.ex[j]);
          break; }
    if (j==i) ndiff++;        /* count different bonds starting with 1 */
    /* TBC changes directed graph into undirected graph! Apr00
     * ww.ex is -1 if sites are exchanged by the program, otherwise +1 */
    w2w[ iw[i][0] ][ iw[i][1] ] = IFww(+ww.ex[i]*) idxnew ;
    w2w[ iw[i][1] ][ iw[i][0] ] = IFww(-ww.ex[i]*) idxnew ;
  }

  if (w1) E1("%d bonds (defined twice) are ignored",w1); 
   
  for(i=0;i<nn;i++){		/* one-site energies */
    idxnew= ndiff + 1; /* avoid 0 */
    for (j=0;j<i;j++)   /* search stored pairs for same */
      if (ww.we[i] == ww.we[j]  /* one-site energies */
       && ww.wu[i] == ww.wu[j]  /* one-site coulomb */
#ifdef HFIELD
       && ww.wf[i] == ww.wf[j]  /* one-site magnetic field hz */
#endif
#ifdef SSANISO
       && ww.wd[i] == ww.wd[j]  /* one-site magnetic field hz */
#endif
      ) { /* found same site */
          idxnew = w2w[ j ][ j ];
          break; }
    if (j==i) ndiff++;        /* count different bonds starting with 1 */
    w2w[ i ][ i ] = idxnew ;
  }
  /* -dt- Fuer den Vergleich zweier Gitter auf topologische Aequivalenz
     ---- waere eine Sortierung der Matrixzeilen (Reindizierung der Plaetze) 
     ---- recht sinnvoll. 
     ---- Sortiere so, dass NZ-Elemente moeglichst dicht an der antidiagonalen
     ---- stehen!
     ---- extra-Programm ../utils/topotest.c (Isomorphie von Graphen)
  */ 
  if(vvv&8){ int k;
    /* get maximum */
    for(i=k=0;i<nn;i++) for(j=0;j<nn;j++) if(k<w2w[i][j]) k=w2w[i][j];
    /* WW-matrix output */
    printf(" Index, WW-matrix (Adjaszensmatrix, max_idx=%d):\n",k);
    for(i=0;i<nn;i++){
      for(j=0;j<nn;j++)printf(((k<256)?((k<16)?"%2x":"%3x"):"%4x"),w2w[i][j]);
      printf("\n");}
  }
  return 0;
}

/* signum for Permutation of N fermions (one per site)
 *  -----------------------------------------------------------------
 *  divide permutations into cycles, return( german: kgV der Zyklenlaengen )
 *  see: Bronstein-Semendjajew, "Taschenbuch der Mathematik"
 *  P2147653=(12)(347)(56) -> 0(1)+2(2)+1(3)+0(4)+0(5)+0(6)+0(7) 
 *                         -> cycl = [0 2 1 0 0 0 0 0 ... ] */
int get_numneg(t_sites pp[NN]){		/* P must be a permutation */
  int cycl[NN],i,j,k;
  for(i=0;i<NN;i++) cycl[i]=0;
  for(j=0;j<nn;j++)			/* hangs, if not a permutation */
    for(k=0,i=pp[j];k<nn;k++,i=pp[i]) 	/* count cycles */
    if(i==j){ cycl[k]++;break; }	/* cycle finished */
  for(i=0,j=1;j<nn;j++)if(j&1)i+=cycl[j]/(j+1);/* count mod2-cycl */
  return i;
}


/* ------------------------------------------------------------- */
/* notation ( 0 1 2 3 ... N ) => ( s0 s1 s2 ... sN ) */
int outlist(t_sites list[NN]){ int j;
  /* printf("\n"); */
  for(j=0;j<nn;j++)printf(" %d",(int)list[j]);
  return 0;} 

/* cycle-notation:  0-1-2-3  4-5  ..., easier to read */
int outcycl(t_sites list[NN]){ int i,j,r; char f[NN]; r=0;
  for(j=0;j<nn;j++) f[j]=1;
  for(j=0;j<nn;j++) for(i=j;i>=0 && i<nn && f[i];i=list[i])
  { r+=printf("%c%d",((j==i)?' ':'-'),i);f[i]=0; }
  return r;}

/*  ----------------------------------------------------------------- */
/*  if ( P3=P1*P2 or P3=P2*P1 ) test=1; else test=0 */
int test_prod(t_sites p1[NN],t_sites p2[NN],
                             t_sites p3[NN]){
  int i;
  for(i=0;i<nn;i++)if(p1[p2[i]]!=p3[i]) break;    
  if(i<nn)
  for(i=0;i<nn;i++)if(p2[p1[i]]!=p3[i]) return 0;
  return 1;
}

/*  -----------------------------------------------------------------
 * !!! is Gray-code more efficient? !!!
 *    gray=bin^bin>>1^bin>>2^...
 *    bin=gray^gray>>1
 *  test if   Pnew=Pold_i1*....*Pold_in   for all combinations i1..in
 *  ret:  0=ok, 0+1: P=1, 1+1: P=P0, 2+1: P=P1, 3+1: P=P0*P1, ...	*/
#if 0
int test_product(t_sites pold[][NN], int num_old, t_sites pnew[NN]){
  unsigned long i,j,k,mask,m;		/* P,Pold must commute */
  t_sites prod[NN];
  k=1l<<num_old;				/* number of masks */
  if(num_old>=8*S_long)E0("Warning num>8long");
  for(mask=0;mask<k;mask++){
    for(j=0;j<nn;j++)prod[j]=pnew[j];		/* prod=Pnew */
    for(i=0,m=1;m<=mask;i++,m<<=1)if(mask & m)	/* select Pold_i */
    for(j=0;j<nn;j++)prod[j]=pold[i][ prod[j] ];/* prod*=Pold_i */ 
    for(j=0;j<nn && prod[j]==(signed int)j;j++);/* prod==1 ? */
    if(j==nn) return (int)mask+1; /* (int) may cause errors !??? */
  }  
  return 0;
}
#else	/* product stored already in table? Apr00 */
int test_product(t_sites pold[][NN], int num_old, t_sites pnew[NN]){
  int i,j;
  for(i=0;i<num_old;i++){
    for(j=0;j<nn;j++)if(pnew[j]!=pold[i][j])break;
    if(j==nn) return 1;
  }  
  return 0;
}
#endif

/*  ----------------------------------------------------------------- */
/*  test if   Pnew=Pold_i   for all 0<=i<num_old */
/*  ret:  0=ok, k+1=twice (pnew==pold[k])	*/
int test_twice(t_sites pold[][NN], int num_old, t_sites pnew[NN]){
  int j,k;
#if 0
  for(k=0;k<num_old;k++){
#else /* faster? wegen ...,P1*P2,P2*P1,... P1,P2 vertauscht! */
  for(k=num_old-1;k>=0;k--){
#endif
    for(j=0;j<nn;j++)if(pold[k][j]!=pnew[j])break;	/* Pnew==Pold_i ? */ 
    if(j==nn) return k+1;
  }  
  return 0;
}

/*  ----------------------------------------------------------------- */
/*  ret:  0=no 1=yes	*/
int test_one(t_sites pnew[NN]){
  int j;
  for(j=0;j<nn;j++)if(j!=pnew[j])return 0;		/* Pnew==1 ? */ 
  return 1;
}

/* ------------------------------------------------
 *  test if [P1,P2]_-==0 respectively P1*P2==P2*P1
 *  ret:  0=commute, 1=not_commuting
 */
int commuting(t_sites p1[NN], t_sites p2[NN]){
  int j;
  for(j=0;j<nn;j++) if( p1[ p2[j] ]!=p2[ p1[j] ] )return 0;
  return 1;
}

/*  ----------------------------------------------------------------- */
/*  test if   [Pnew,Pold_i]_-==0 bzw. P*Pold_i==Pold_i*P  */
/*  ret:  0=ok else i+1		*/
int test_commute(t_sites pold[][NN], int num_old, t_sites pnew[NN]){
  int i;
  for(i=0;i<num_old;i++) if( !commuting( pold[i], pnew ) ) return i+1;
  return 0;
}


/*  ----------------------------------------------------------------- */
/*  test if permutation; ret: 0=ok, else no permutation */
int test_permutation(t_sites pp[NN]){	/* static int, aufrufe zaehlen ??? */
  int i,j;
  for(i=0;i<nn;i++){	/* jede Zahl i=0..NN-1 muss genau 1 mal vorkommen */
    for(j=0;j<nn;j++)if(pp[i]==j)break;		/* zahl j gefunden */
    if(j==nn)return i+1; }			/* nicht gefunden */
  return 0;
}

/*  test if s1-collision; ret: 0=ok, else LM-conflict */
int test_s1collision(t_sites pp[NN]){
  int i;
  for(i=0;i<nn;i++)
    if(pp[i]!=i && sym.lm[i]==sym.lm[pp[i]]) return -1;
  return 0;
}

int test_all_sym(){		/* implementiert um array-ov zu testen */
  int i; printf(" test_all_sym");
  for(i=0;i<sym.numsym;i++) if( test_permutation(sym.i[i]) ) break;
  if( i<sym.numsym ){ E1("P%d isn't a permutation",i); return 1;}
  return 0;
}


/*  ----------------------------------------------------------------- */
/*  ret: n_minimum for P^n==1  fastest way: P=(12)(347)(56) -> 2*3=6 */
int test_cycles(t_sites pp[NN]){		/* P must be a permutation */
  int cycl[NN]; int i,j,k;		/* Jan99 signed char => int */
  for(i=0;i<nn;i++) cycl[i]=0;
  for(j=0;j<nn;j++)			/* hangs, if not a permutation */
    for(k=0,i=pp[j];k<nn;k++,i=pp[i]) 	/* count cycles */
    if(i==j){ cycl[k]++;break; }	/* cycl finished */
#if 0 /* wrong version ? */
  for(i=j=1;j<nn;j++)if(cycl[j])i*=(j+1);/* multiply cycl-lengths */
#else
  for(i=1,j=nn-1;j>0;j--)if(cycl[j])if(i%(j+1)!=0)i*=(j+1);/* kgV */
#endif
  return i;
}

/*  ----------------------------------------------------------------- */
/*  Vertauscht H mit P?	ret: 0=ok, [P,H]==0, PH==HP; !0: [P,H]!=0	 */
int test_symmetry(t_sites pp[NN]){		/* P must be a permutation */
  int i,j;	
  ini_ww();
  for(i=0;i<nn;i++)for(j=i;j<nn;j++)
    if(w2w[i][j]!=w2w[ pp[i] ][ pp[j] ])return i*nn+j+1;
  return 0; }


/*  ----------------------------------------------------------------- */
/*  test all things on Pnew, output on errors   */
/*  ret:  0=ok 		*/
int test_all(t_sites pold[][NN], int num_old, t_sites pnew[NN]){
  int i;
  printf("\n test_sym  ");outcycl(pnew);printf(" i=%u",num_old);
  i=test_permutation(pnew); if(i){	
  	printf("\n P%u is not a permutation (%u not found)\n",num_old,i-1); 
  	return -1; }
  i=test_symmetry(pnew); if(i){
  	printf(" P%u isn't a symmetry (%u-%u)\n",num_old,(i-1)/nn,(i-1)%nn);
  	return -2; }
  i=test_cycles(pnew); printf(" l=%d",i);
        /* if(i!=2){ printf(" l!=2\n"); return -3; } */
  i=test_commute(pold,num_old,pnew); if(i){
  	printf(" [P%u,P%u]!=0",num_old,i-1);
#ifdef CONFIG_ABEL
  	return i;
#endif  	 
      }
  i=test_product(pold,num_old,pnew); if(i){	
  	printf(" P%u==Prod_(i in %u) P_i\n",num_old,i-1);
  	return nn+i; }
#ifdef CONFIG_S1SYM		/* Nov 2001 v1.8.1 */
  i=test_s1collision(pnew); if(i) { printf(" sym-s1-collision"); return -4; }
#endif
  printf(" OK %d",get_numneg(pnew));	
  return 0;
}

/*  -------------------------------------------------------
 *  add permutation to list (sym) if check==ok
 *  return 1..7 otherwise (print if v!=0 (generator) or error)
 */
int addperm(int *num_sym, t_sites s1[NN], int k, int v, int *k2 ){
  int j,l;
  l=test_cycles(s1); while(k>=l) k-=l;
  if(test_one(s1))		    return 2;	/* P=1 */
  if(*num_sym>=MaxSym)              return 3;	/* not enough space */
  *k2=test_twice(sym.i,*num_sym,s1);		/* already stored */
  if(*k2){ if(sym.k[*k2-1]!=k)      return 4;   /* .. and other k */
           else			    return 1; } /* .. but same k */
  for(j=0;j<nn;j++) sym.i[*num_sym][j]=s1[j];
  sym.l[*num_sym]=l;			/* kgV der cyclenlaengen */
  sym.k[*num_sym]=k;			/* k=k1+k2 if P=P1*P2 (=P2*P1) */
  sym.g[*num_sym]=v&1;			/* generator of sym-group */
#if VecType & 4
  sym.w[*num_sym]=C(cos(k*2*M_PI/l),sin(k*2*M_PI/l));	/* complex */
#else
  sym.w[*num_sym]= (2*k==l) ? -1 : (k==0) ? +1 : 0 ;	/* else complex */
#endif
#if (VecType & 4) == 0 && defined(SBase)	/* use complex vectors */
  if(k!=0 && 2*k!=l){
    E0("unsupported k, please switch to complex mode (VecType|=4)");
    printf(", l=%d k=%d i=%u",l,k,*num_sym);
    return 5;					/* ... no complex */ 
  }
#endif
  (*num_sym)++;			/* ok, added */
  return 0; }  

/* --- product P3 of two permutations P1,P2 and its expectation value w3
   --- w3 = w1*w2 = exp(2*pi*I*k1/l1)*exp(2*pi*I*k2/l2)
   ---    = exp(2*pi*I*(k1/l1+k2/l2))
   ---    = exp(2*pi*I*(k1*(l3/l1)+k2*(l3/l2))/l3)  (l3%l1==0,l3%l2==0)
   --- k3 = k1*(l3/l1)+k2*(l3/l2)			1998-166 JS
   ---------------------------------------------------------------- */
int sum_k( int l12,  int k1, int l1,  int k2, int l2 ){
  if   ( (l12*(l2*k1+k2*l1))%(l1*l2) ) return -1;
  return (l12*(l2*k1+k2*l1))/(l1*l2); }	/* addiere 2 Brueche */

/* Feb99 JS, multiply two Permutations P=P1*P2 */
t_sites *mulP( t_sites *p, t_sites *p1, t_sites *p2, int num){
  int j; for(j=0;j<num;j++) p[j]=p1[ p2[j] ];
  return p;
}

/* insert perm s1 and all products s1*sym.i[], if simultaneously measurable */
/* return: 0=ok 1=error (not stored), v&1 generator, v&2 output */
int addprod( int *num_sym, t_sites s1[NN], int ki, int v ) {
  t_sites p1[NN];			/* product permutation */
  int i2,i3,j,l,kl,rc=0,err,k2;
  i3=*num_sym;				/* do not use ki<0 */
  if( ki>=0 && !(rc=addperm(num_sym, s1, ki, v, &k2)) ){
    if(v&2){ printf("\n l=%2d k=%2d ",sym.l[i3],sym.k[i3]); outcycl(s1); }
    /* --- will take more time! until num_sym^2 (MaxSym^2) tests --- */
    for(i2=i3;i2<*num_sym && rc==0;i2++)	/* also all new products */
    for(l=0,i2=*num_sym-1;l<=i2 && rc==0;l++)	/* build products */
    for(j=0;j<2 && rc==0;j++) {
      if( j==0 ) mulP(p1,sym.i[i2],sym.i[l],nn);	/* P1=S1*Sl */
      else       mulP(p1,sym.i[l],sym.i[i2],nn);	/* P1=Sl*S1 */
      if( j==1 && commuting(sym.i[l],sym.i[i2]) ) break;
      /* new k */   
      kl=test_cycles(p1); kl=sum_k(kl,sym.k[i2],sym.l[i2],sym.k[l],sym.l[l]);
      if(kl<0){
        if(v&2)printf(" k/%d=%d/%d+%d/%d",
          test_cycles(p1),sym.k[i2],sym.l[i2],sym.k[l],sym.l[l]);
        rc=6;break; }				/* incompatible */
      err=addperm(num_sym,p1,kl,0,&k2); 
      if(err==1 || err==4){ /* GenOld * GenNew already stored ??? */
        if( sym.g[l] && sym.g[i2] ) /* && sym.l[i2]>sym.l[l] ) */
        printf("\n Warning: P%d depence from P%d and P%d, %d",l,k2-1,i2,*num_sym);
      }
      if(err==3){ rc=3; printf(" symtable full"); }
      if(err==4){ rc=err;   break; }	/* kollision */
      if(err >3){ rc=10+err;break; }	/* other error */
    }
    if(!rc)if(v&2)printf(" i=%d...%d",i3,*num_sym-1);
  }
  if(rc>2){ *num_sym=i3; if(vvv)if(v&2)printf(" err=%d",rc); }
  return rc;
}

/* add perm s1 and all products s1*sym.i[], if simultaniously measurable */
/* return first possible k */
int all_k( int num_sym, t_sites s1[NN] ){
  int ki,i3,ls1,rc=-1;
  ls1=test_cycles(s1);
  i3=num_sym;
  for(ki=0;ki<ls1;ki++)		/* Nov99 some times we need kx=1/5,ky=4/5 */
#if (VecType & 4)==0
  if((2*ki)%ls1==0)		/* only real phases */
#endif  
  {
    num_sym=i3;	/* restore */
    if(!addprod(&num_sym,s1,ki,1)){
      if(rc<0) { rc=ki; printf(" possible k="); }
      printf(" %d",ki);
    }
  }
  if(rc<0) printf(" no k possible");
  return rc; }

/* --- output set of permutations which commute with H --- */
int save_syms( int num_sym ){
 FILE *f1; int j,j2;
 if(vvv){ printf(" write %s",FILE_S2); }
 f1=fopen(FILE_S2,"w"); if( !f1 ) E0("save_syms: fopen failed");
 if(f1)for(j2=0;j2<num_sym;j2++){
     if(j2==0)fprintf(f1," %u\n",num_sym); 
     for(j=0;j<nn;j++)fprintf(f1," %u",sym.i[j2][j]);
     fprintf(f1," l=%u k=%u %4d\n",sym.l[j2],sym.k[j2],j2); } 
 if(f1)fclose(f1);
 return 0;
}

/* ---- input: w2w;  out: sym.lm+li+ls, sym.num_lm ---- */
int ini_sym_lm(){
 int i,j1,j2,l;
 sym.num_lm=nn;					/* counter */
 for(i=0;i<nn;i++){ sym.lm[i]=i; sym.li[i]=0; sym.ls[i]=1; sym.ln[i]=nn; } 
#if 1                  /* --- search for Lieb-Mattis-Cluster --- */  
 for(j1=0;j1<nn-1;j1++)if(sym.lm[j1]==j1){ i=0;	/* site j1 == ..   */
   for(j2=j1+1;j2<nn;j2++)if(sym.lm[j2]==j2){	/*      .. == j2 ? */
     if( w2w[j2][j2]!=w2w[j1][j1]
      || w2w[j1][j2]!=w2w[j2][j1] ) continue;	/* same neighbours? */
     for(l=0;l<nn;l++)
     if( l!=j1 && l!=j2 && w2w[j1][l]!=w2w[j2][l] ) break;
     if( l<nn ) continue;			/* broken for() */
     sym.lm[j2]=sym.lm[j1];
     sym.li[j2]=++i;
     sym.num_lm--;				/* LM-coupled */
   }  
 }
 for (j1=0;j1<nn;j1++) { /* get LM size == number of members */
   for (j2=0;j2<nn;j2++)
     if (j1!=j2 && sym.lm[j1]==sym.lm[j2]) { /* other member found */
         sym.ls[j1]++; /* increase spinlength */
         if (sym.ln[j1]>=nn && j2>j1) sym.ln[j1]=j2;  /* next member v2.21 */
     }
 }
 printf(  " LS=%u ",sym.num_lm); outlist(sym.ls); /* size of composed spin */
 printf("\n LI=%u ",sym.num_lm); outlist(sym.li); /* index member */
 printf("\n LM=%u ",sym.num_lm); outlist(sym.lm); /* index composed spin */
 printf("\n LN=%u ",sym.num_lm); outlist(sym.ln); /* next member */
#endif 
 return 0;
}

/* ---------------------------------------------------------
 *    read sym-file
 *    out: num_sym, sym.xxx
 *
 * test before trusting !!! erst -> s1, testen, k-load, addperm
 * or without defaults (read in)
 */
int read_syms( int num_sym, int *ki, int *igen ){
 t_sites s1[NN],p1[NN];		/* permutation of sites, for testing */
 char   buf[1024],*buf2;	/* enough? */
 int i,j,j2,k,rc;
 FILE *f1;
 f1=fopen(FILE_S1,"r");	/* try to load old symmetries */	
 if(f1){
    fgets(buf,1024,f1);		/* read line Apr99 */
    /* sscanf(buf," %u",&i); # changed on 1.7.12+ */
    i=(int)strtol(buf,NULL,10);
    if(vvv)printf("\n read %d " FILE_S1,i);
    for(k=0;k<i;k++){
     fgets(buf,1024,f1);buf2=buf;	/* read line Apr99 */
     for(rc=j=0;j<nn;j++) if( (p1[j]=(int)strtol(buf2,&buf2,10))<0 ) rc=1;
     if( rc ){  /* cycle form - converting */
       /* if(vvv)printf("\n cyclic form detected "); */
       for( j=0; j<nn; j++ ){ s1[j]=p1[j];p1[j]=j; }
       for( j2=j=0; j<nn; j++ ){
         if( abs(s1[j]>=nn) ) F0("index>NN");
         if( s1[j]>=0 ) { j2=s1[j]; if( p1[ j2 ]!=j2 )E0("wrong data"); }
         else           { p1[ abs(s1[j-1]) ] = abs(s1[j]);
                          p1[ abs(s1[j  ]) ] = j2;
                        }
       }
       /* if(vvv)for(j=0;j<nn;j++) printf(" %d",p1[j]); */
     }
     if( test_all(sym.i,num_sym,p1) ) continue; /* break; v1.7.12+ */
     if(*ki<0) (*ki)++;	                        /* skip (-ki) P's! */
     if(*ki>=0){
       *ki=sym.kgen[*igen];(*igen)++;		/* input defaults */
       addprod(&num_sym,p1,*ki,3);
     }
   } fclose(f1);
 }
 return num_sym;
}
  
/* ------------------------------------------------------------------------
 * recursive generation of permutations v1.7.13+ (very fast and elegant)
 * how to generate _unique_ permutations?
 * using sym_pairs instead of w2w would be faster??? 
 * better use neighbour instead of next lower spin number ???
 * I do not know why I could not write this for earlier versions.
 * It was so easy at last (2h progging).
 * - only ii is changed during recursion
 * ex: pyrochlore_256sites takes > 14days/P2GHz, T0=4min
 *    send Signal to skip search? m*SIGUSR2+SIGUSR1 => volatile int sig1=m;
 *    return 0=ok 1=abort 
 */
extern volatile int sig1;
int get_recursive_perm( t_sites s1[NN],
  int *num_sym, ULong *k, int *ki, int *igen, int ii ){
  t_sites s2;
  int rc,i,j;
  static unsigned long recursive_calls=0; /* measure recursiv-cycles */
  if (sig1) return 1;
  if ( ii<nn ){  /* s1 not complete */
    /* loop i=ii+1,...,nn-1,0,1,...,ii would generate translations before reflections!? */
    s2=s1[ii]; /* save old state */
    for( i=0; i<nn; i++ ) if( w2w[i][i]==w2w[ii][ii] ){
      s1[ii]=i; /* modify state */
      for( j=0; j<ii; j++ ) /* test if i is already used and for symmetry */
        if( s1[j]==i || w2w[ii][j]!=w2w[ i ][ s1[j] ] ) break;
      if( j<ii ) continue;
#if CONFIG_S1SYM	/* keine LM-Permutationen new (schneller!?) */
      if(sym.num_lm<nn){
        if(i!=ii && sym.lm[ii]==sym.lm[i]) continue;
        if(i!=ii && sym.li[ii]!=sym.li[i]) continue;
        for( j=0; j<ii; j++)
          if(sym.lm[ii]==sym.lm[j] && sym.lm[s1[ii]]!=sym.lm[s1[j]])
          break;
        if(j<ii) continue;
      }
#endif
      recursive_calls++;
      if ( get_recursive_perm( s1, num_sym, k, ki, igen, ii+1 ) ) return 1;
    } s1[ii]=s2; /* restore state */
  } else {      /* s1 complete => store */
    if (vvv&8) {
      printf(" rcalls=%lu\n test permutation: ",recursive_calls); 
      outcycl(s1);
    }
#ifndef CONFIG_NOABEL
    if(test_commute(sym.i,*num_sym,s1)) return 0;	/* vorlaeufig */
#endif
    (*k)++;
    if(*num_sym>=MaxSym) return 0;                 if (vvv&8) printf(" .");
    if( test_one(s1) ) return 0;                   if (vvv&8) printf(".");
    if( test_twice(sym.i,*num_sym,s1) ) return 0;  if (vvv&8) printf(".");
    if(*ki<0) (*ki)++;	/* (-ki) Ps ueberspringen! */
    if(*ki>=0){
      *ki=sym.kgen[*igen];(*igen)++;		/* input vorgaben */
    }
    rc=addprod( num_sym, s1, *ki, 3 );
    if (vvv&8) printf(" rc=%d",rc);
    /* rc=0 means addperm=ok, rc=3: full memory (ex: octahedron S=1) */
    if(rc!=0 && rc!=3 && rc!=2) all_k( *num_sym, s1);
  }
  return 0;
}

/*  ----------------------------------------------------------------- 
 * #################### version 5 ############################
 *    ( 1 2 3 4 5 6 7 ... )     rekursiv v1.7.13+ (fast)
 * s1=( 3 4 1 2 6 5 7 ... )  -> Permutiere
 * test & create symmetries/permutations, return 0=ok
 */
int symcreate5(int maxsym){ 
 t_sites s1[NN];
 int num_sym=0,ki=0,i,	/* ki=QuantumNumber */
     igen=0;		/* index to sym.kgen */
 ULong  k=0l;			/* k= number of (P^2=1) Symmetries found */

 sym.numsym=sym.numall=num_sym=0; if(maxsym==0) return 0;

 num_sym=read_syms(num_sym, &ki, &igen);

 if(vvv&1)printf(" +++ search T%u..",num_sym);
 
 for( i=0; i<nn; i++ ) s1[i]=-1; /* initialize, not necessary??? */
 if (vvv) printf("\nsym-search can be aborted by sending SIGUSR2+SIGUSR1 ");
 if (get_recursive_perm( s1, &num_sym, &k, &ki, &igen, 0 ))
   printf("\n recursive_perm was aborted by SIGUSR2+1");
 
 if(vvv&9)printf("\n sym_stored= %u",num_sym+1); /* incl. P=1 */
 if(vvv&9)printf(" sym_found= %lu",(unsigned long)k);
#ifdef CONFIG_ABEL 
 if(num_sym!=maxsym){ printf(" Warning: %d!=Sym(%d) !!!\n",num_sym,maxsym); }
#endif

 sym.numsym=num_sym;		/* max. Sym */
 sym.numall=num_sym;	/* was ist mit Sud ? */
 if( (int)k<num_sym ){ E1("k<num_sym ??? k=%d",(int)k); }
 /* 2*k ??? */
 if(  k+512<MaxSym){ printf("\n reduce  MaxSym to %d to spare memory!",(int)k+2); }
 if(  k    >MaxSym){ printf("\n enlarge MaxSym to %d to be faster!",(int)k+1); }
 return 0;
}

/*
      idx=1,2,0 iw=1-2 2-0 0-1 pos=0 1 2
                iw=0-2 0-1 1-2 pos=1 2 0
 */ 
int reindex(){		/* reindiziere sites fuer sym-lintab => AAAABBBB */
  int i,j,idx[NN],iw2[Nw][2];double po2[NN][3];
  for(i=0;i<nn;i++){ idx[i]=i; for(j=0;j<3;j++)po2[i][j]=pos[i][j]; }
  for(i=0;i<nw;i++){ for(j=0;j<2;j++)iw2[i][j]=iw[i][j]; }
  printf(" re-idx=");
  /* suche guenstige Form */
#if 0 /* Lin unsym -> sym-idx ohne Anwendung von Symmtrie geht nicht, sh.t */
  /* ABABABAB => AAAABBBB */
  if((iw[0][0]^iw[0][1])&1)
  for(i=0;i<nn;i++){ idx[i]=(i&1)?i/2+nn/2:i/2; } 
#endif
  /* reindiziere */
  for(i=0;i<nw;i++){
    iw[i][0]=idx[ iw2[i][0] ];
    iw[i][1]=idx[ iw2[i][1] ];
    if(iw[i][0]>iw[i][1]){ j=iw[i][0];iw[i][0]=iw[i][1];iw[i][1]=j; } }
  for(i=0;i<nn;i++){ printf(" %d",idx[i]);
    for(j=0;j<3;j++)pos[i][j]=po2[ idx[i] ][j];
  }
  return 0;}

/* ----- check for symmetric correlations (inp=sym.xx) -----
 * if symmetries of the system are known, some one- and two-particle
 * correlations are identical and we could save processor time by
 * calculating only one of this identical correlations,
 * - for correct spin.chk3(), sym.korr must be sorted! 
 */
int symkorrelation(){
  int i,j1,j2,i1,i2,i3,n=0,korr2[NN][NN],sl[NN];
  for(i2=0;i2<nn;i2++)for(i1=0;i1<nn;i1++)korr2[i2][i1]=-1;  /* init as not set */
  /* ToDo:
   *    correlation table only set if syms are used,
   *    better would be if get_recursive_perm would would
   *    generate this matrix in every case (not k-dependend)
   */
  for(i2=0;i2<nn;i2++)for(i1=i2;i1<nn;i1++)if(korr2[i2][i1]<0){
    for(i=-1;i<sym.numall;i++){		/* i=-1 fuer P=1 - Symmetrie */
      j1=((i<0)?i1:sym.i[i][i1]); 
      j2=((i<0)?i2:sym.i[i][i2]);
#if 1 /* new method using sym.lm */
      for(j1=0;j1<nn;j1++)for(j2=0;j2<nn;j2++)   /* not very efficient */
      if(sym.lm[j1]==sym.lm[ (i<0)?i1:sym.i[i][i1] ]
      && sym.lm[j2]==sym.lm[ (i<0)?i2:sym.i[i][i2] ])
      if((j1==j2 && i1==i2) || (j1!=j2 && i1!=i2))
#endif
      { if(korr2[j2][j1]<0){ korr2[j1][j2]=korr2[j2][j1]=n; }
        else if(korr2[j2][j1]!=n){ F0(" ???"); } }
    }
    n++; 
  } 
  printf("\ndifferent pairs: ndp= %d\n weight: wdp=",n);
  sym.num_dkorr=n;
  for(i3=0;i3<n;i3++){ j1=0;
    for(i2=0;i2<nn;i2++)for(i1=0;i1<nn;i1++)if( korr2[i2][i1]==i3 )j1++;
    if(i3 <20)printf(" %d",j1);
    if(i3==20)printf(" ...");   /* awk does not like long lines 04-00 */
  }
  for(i2=0;i2<nn;i2++){
    if(vvv&8)printf("\n");
    for(i1=0;i1<nn;i1++){
      if(vvv&8)printf(((n<11)?" %1d":" %2d"),korr2[i2][i1]); 
      sym.korr[i2][i1]=korr2[i2][i1];
      /* fuer einfache paarkorrel. <ZiZj> bereitstellen */
    }
  }
  /* create a table for symmetricaly different pairs,
   * used in other subroutines */
  for(i3=-1,i1=0;i1<nn;i1++)for(i2=0;i2<nn;i2++)	/* Jan 00 */
  if(sym.korr[i1][i2]>i3){	/* only different correlations */
    i3=sym.korr[i1][i2]; if( i3>=nn*nn )ERR(9000);
    sym.dkorr[i3][0]=i1;
    sym.dkorr[i3][1]=i2;
  }
  /* get SUBLATTICES v1.53 
   */
  if (sym.subl>1) { printf(" skip SL autodetection\n");return 0;}
  for(i2=0;i2<nn;i2++)sym.sl[i2]=0;sym.subl=1;
  for(i1=1;i1<sym.num_dkorr;i1++){	/* korrelation-type */
    for(i2=0;i2<nn;i2++){	/* an jedem Platz vorhanden ? */
      for(i3=0;i3<nn;i3++)if(sym.korr[i2][i3]==i1)break; /* ja */
      if( i3==nn ) break; /* nein */
    }
    if( i2<nn ) { /* printf(" not every site connected"); */
      continue;} /* nein */
    for(i2=0;i2<nn;i2++)sl[i2]=i2;
    for(j1=1;j1;)
    for(j1=i2=0;i2<nn;i2++){	/* set connected sites equal */ 
      for(i3=i2+1;i3<nn;i3++)
      if(sym.korr[i2][i3]==i1 && sl[i3]!=sl[i2]){ 
        if(sl[i3]>sl[i2]) sl[i3]=sl[i2];
        else              sl[i2]=sl[i3];
        j1=1;
      } 
    }
    j1=-1;	/* number SL */
    for(i2=0;i2<nn;i2++){
      if(sl[i2]>j1){ j1++;
        for(i3=nn-1;i3>=i2;i3--)if(sl[i3]==sl[i2])sl[i3]=j1;
      }
    }
    printf("\n sublattice %2d %2d :",i1,j1+1);
    /* found!!! */
    for(i2=0;i2<nn;i2++) printf(" %d",sl[i2]);
    /* set bipartit as default! ToDo: better bibipartit ??? or code bits */ 
    if(j1==1){ sym.subl=j1+1;for(i2=0;i2<nn;i2++) sym.sl[i2]=sl[i2]; }
  } 
  printf("\n");
  return 0; }

int ini_inverse(){
  int i1,i,j;
  for(i1=0;i1<sym.numall;i1++)sym.inverse[i1]=~0;	/* mark as not set */
  for(i1=0;i1<sym.numall;i1++)if(sym.inverse[i1]==~0){	/* test not set */
    for(i=i1;i<sym.numall;i++){
      for(j=0;j<nn;j++)if(sym.i[i][ sym.i[i1][j] ]!=j)break;
      if(j==nn){ sym.inverse[i1]=i;sym.inverse[i]=i1; }	/* inverse found */
    }
    if(sym.inverse[i1]==~0)F1(" inverse SymOp P%d^(-1) not found",i1)
  }
  return 0;}

/* --- compare products (used for commutations, classes etc)
   ---  ret=1:  p1*p2==p3*p4    --- */
int comp1234(t_sites *p1, t_sites *p2,
             t_sites *p3, t_sites *p4){ int i;
  for(i=0;i<nn;i++) if( p1[ p2[i] ] != p3[ p4[i] ] ) return 0;
  return 1; }

/* get "Aequivalenzklassen", A and B are equivalent if A*C=C*B (for all C)
 * bei Permutationen gehoeren alle Elemente mit gleicher Zyklendarstellung
 * zur selben Klasse (schneller!!!???), (liste der Zyklenlaengen)
 */
int ini_aclass(){
  int i1,i2,i,j,anr=-1,sym_aclass[MaxSym];
  if(~vvv&8) return 0;			/* takes long time for numsym>1000 */
  for(i1=0;i1<sym.numall;i1++)sym_aclass[i1]=~0;	/* mark as not set */
  for(i1=0;i1<sym.numall;i1++)if(sym_aclass[i1]==~0){	/* test not set */
    anr++; sym_aclass[i1]=anr; 				/* new aeq-class */
    printf(" C%d: %d",anr,i1);
    for(i2=i1+1;i2<sym.numall;i2++)if(sym_aclass[i2]==~0){ /* Pi1*Pi=Pi*Pi2 */
      for(i=0;i<sym.numall;i++){
        for(j=0;j<nn;j++)
          if(sym.i[i1][ sym.i[i][j] ]!=sym.i[i][ sym.i[i2][j] ])break;
        if(j==nn){ 					/* same aclass */
          sym_aclass[i2]=sym_aclass[i1];
          printf(" %d",i2);
          break; }
      }
    }
    printf("\n"); 
  }
  anr++;printf(" c=%d Aequivalenzklassen, Vertauschungen:\n",anr);
  /* T=1 nicht mitgezaehlt!? */
  /* pruefe, welche Klassen mit welchen anderen vertauschen */
  for(i1=0;i1<anr;i1++){
    printf(" class %2d",i1);
    /* suche mitglied der Klasse i1 */
    for(i=0;i<sym.numall;i++) if(sym_aclass[i]==i1) break;
    printf(" invClass %2d",sym_aclass[sym.inverse[i]]);
    printf(" cycles %2d\n",sym.l[i]);
  }
  /* pruefe, welche syms mit welchen anderen vertauschen */
  i2=sym.numall;if(i2>40) i2=40;	/* begrenze Ausgaben */
  for(i=0;i<i2;i++)if(i<=sym.inverse[i]){	/* most compact */
    printf(" P %2d:",i);
    for(j=0;j<i2;j++)if(j<=sym.inverse[j])
    printf(" %c",((comp1234(sym.i[i],sym.i[j],sym.i[j],sym.i[i]))?'*':'-'));
    printf(" cycles %2d  class %2d  inv %2d k %2d\n",
      sym.l[i],sym_aclass[i],sym.inverse[i],sym.k[i]);
  }
  for(i=0;i<i2;i++)if(i<=sym.inverse[i]){
    printf(" P %2d:",i); outcycl(sym.i[i]);
    printf(" inv %d\n",sym.inverse[i]);
  }
  return 0;
}

int symcreate(int maxsym){ 
  int i; /* system("ps $PPID"); */
#ifdef CONFIG_ABEL
  if(vvv)printf(" set ABEL\n");
#endif
  reindex();
  ini_ww();			/* initialize w2w */
  ini_sym_lm();
  i=symcreate5(maxsym);	/* new code */
  printf(" numsym= %d symall= %d",sym.numsym+1,sym.numall+1); /* incl. P=1 */
  save_syms( sym.numsym );
  symkorrelation();		/* check for symmetric korrelations */
  ini_inverse();
  ini_aclass();
  return i; 
}
     
